"""
Author: RedFantom
License: GNU GPLv3
Copyright (c) 2017-2018 RedFantom
"""
from ._utils import is_python_3
from ._widget import ThemedWidget
if is_python_3():
    import tkinter as tk
else:
    import Tkinter as tk


class ThemedTk(tk.Tk, ThemedWidget):
    """
    Tk child class that supports the themes supplied in this package.
    A theme can be set upon initialization or during runtime. Can be
    used as a drop-in replacement for the normal Tk class.
    """
    def __init__(self, *args, **kwargs):
        """
        :param theme: Theme to set upon initialization. If theme is not
                      available, will fails silently.
        """
        theme = kwargs.pop("theme", None)
        # Initialize as tk.Tk
        tk.Tk.__init__(self, *args, **kwargs)
        # Initialize as ThemedWidget
        ThemedWidget.__init__(self, self.tk)
        # Set initial theme
        if theme is not None and theme in self.get_themes():
            self.set_theme(theme)
        return



"""
Author: RedFantom
License: GNU GPLv3
Copyright (c) 2017-2018 RedFantom
"""
import sys
from ._widget import ThemedWidget
if sys.version_info.major is 2:
    import ttk
else:
    from tkinter import ttk


class ThemedStyle(ttk.Style, ThemedWidget):
    """
    Style that supports setting the theme for a Tk instance. Can be
    used as a drop-in replacement for normal ttk.Style instances.
    Supports the themes provided by this package.
    """
    def __init__(self, *args, **kwargs):
        """
        :param theme: Theme to set up initialization completion. If the
                      theme is not available, fails silently.
        """
        theme = kwargs.pop("theme", None)
        # Initialize as ttk.Style
        ttk.Style.__init__(self, *args, **kwargs)
        # Initialize as ThemedObject
        ThemedWidget.__init__(self, self.tk)
        # Set the initial theme
        if theme is not None and theme in self.get_themes():
            self.set_theme(theme)
        return

    def theme_use(self, theme_name=None):
        """
        Set a new theme to use or return current theme name
        :param theme_name: name of theme to use
        :return: active theme name
        """
        if theme_name is not None:
            self.set_theme(theme_name)
        return ttk.Style.theme_use(self)

    def theme_names(self):
        """
        Alias of get_themes() to allow for a drop-in replacement of the
        normal ttk.Style instance.
        :return: Result of get_themes()
        """
        return self.get_themes()



"""
Author: RedFantom
License: GNU GPLv3
Copyright (c) 2017-2018 RedFantom
"""


def shift_hue(image, hue):
    """
    Shifts the hue of an image in HSV format.
    :param image: PIL Image to perform operation on
    :param hue: value between 0 and 2.0
    """
    hue = (hue - 1.0) * 180
    img = image.copy().convert("HSV")
    pixels = img.load()
    for i in range(img.width):
        for j in range(img.height):
            h, s, v = pixels[i, j]
            h = abs(int(h + hue))
            if h > 255:
                h -= 255
            pixels[i, j] = (h, s, v)
    return img.convert("RGBA")


def make_transparent(image):
    """Turn all black pixels in an image into transparent ones"""
    data = image.copy().getdata()
    modified = []
    for item in data:
        if _check_pixel(item) is True:
            modified.append((255, 255, 255, 255))  # White transparent pixel
            continue
        modified.append(item)
    image.putdata(modified)
    return image


def _check_pixel(tup):
    """Check if a pixel is black, supports RGBA"""
    return tup[0] == 0 and tup[1] == 0 and tup[2] == 0



"""
Author: RedFantom
License: GNU GPLv3
Copyright (c) 2017-2018 RedFantom
"""
# Standard library
import os
from shutil import copytree, rmtree, copyfile
# Packages
from PIL import Image, ImageEnhance
import ttkthemes._imgops as imgops
# Own modules
from ttkthemes import _utils as utils


class ThemedWidget(object):
    """
    Provides functions to manipulate themes in order to reduce code
    duplication in the ThemedTk and ThemedStyle classes.
    """

    pixmap_themes = [
        "arc",
        "blue",
        "clearlooks",
        "elegance",
        "kroc",
        "plastik",
        "radiance",
        "winxpblue"
    ]

    def __init__(self, tk_interpreter):
        """
        Loads themes into tk interpreter
        :param tk_interpreter: tk interpreter for tk.Widget that is
                               being initialized as ThemedWidget
        """
        self.tk = tk_interpreter

        # Change working directory temporarily to allow Tcl to work
        with utils.temporary_chdir(utils.get_file_directory()):
            # Load the themes
            self.folder = os.path.dirname(os.path.realpath(__file__)).replace("\\", "/")
            self.tk.call("lappend", "auto_path", "[%s]" % self.folder + "/themes")
            self._img_support = self._load_tkimg()
            self.tk.eval("source themes/pkgIndex.tcl")

    def set_theme(self, theme_name):
        """
        Set new theme to use. Uses a direct tk call to allow usage
        of the themes supplied with this package.
        :param theme_name: name of theme to activate
        """
        self.tk.call("package", "require", "ttkthemes")
        self.tk.call("ttk::setTheme", theme_name)

    def get_themes(self):
        """Return a list of names of available themes"""
        self.tk.call("package", "require", "ttkthemes")
        return list(self.tk.call("ttk::themes"))

    @property
    def themes(self):
        """Property alias of get_themes()"""
        return self.get_themes()

    def set_theme_advanced(self, theme_name, brightness=1.0,
                           saturation=1.0, hue=1.0,
                           preserve_transparency=True, output_dir=None,
                           advanced_name="advanced"):
        """
        Load an advanced theme that is customized upon runtime. Requires
        write permission into a temporary directory that is used to
        temporarily store the custom theme. Only available with Img
        library support (Tk 8.6 or higher, Python 3 unless you have
        compiled Tk yourself).
        :param theme_name: pixmap theme name to base the new theme on
        :param brightness: brightness modifier for the pixmaps
        :param saturation: saturation modifier for the pixmaps
        :param hue: hue modifier for the pixmaps, limited to 0.0-2.0
        :param preserve_transparency: When True, all black pixels in the
            resulting images will be set to transparent instead, as
            transparency is lost during RGBA->HSV conversion. Has no
            effect when not changing hue.
        :param output_dir: directory to put the theme in. By default,
            a volatile temporary directory is used
        :param advanced_name: Name to give to the new theme
        :raises: RuntimeError when PNG images are not supported by Tk.
            This usually applies to Python 2 platforms.
        :raises: ValueError is theme is not available or valid
        :raises: RuntimeError when the same name is used twice
        """
        if not self.img_support:
            raise RuntimeError("PNG images are not supported by your Tkinter installation")
        # Check if the theme is a pixmap theme
        if theme_name not in self.pixmap_themes:
            raise ValueError("Theme is not a valid pixmap theme")
        # Check if theme is available in the first place
        if theme_name not in self.themes:
            raise ValueError("Theme to create new theme from is not available: {}".format(theme_name))
        if advanced_name in self.themes:
            raise RuntimeError("The same name for an advanced theme cannot be used twice")
        # Unload advanced if already loaded
        output_dir = os.path.join(utils.get_temp_directory(), advanced_name) if output_dir is None else output_dir
        self._setup_advanced_theme(theme_name, output_dir, advanced_name)
        # Perform image operations
        image_directory = os.path.join(output_dir, advanced_name, advanced_name)
        self._setup_images(image_directory, brightness, saturation, hue, preserve_transparency)
        # Load the new theme
        with utils.temporary_chdir(output_dir):
            self.tk.call("lappend", "auto_path", "[{}]".format(output_dir))
            self.tk.eval("source pkgIndex.tcl")
            self.set_theme(advanced_name)

    @staticmethod
    def _setup_advanced_theme(theme_name, output_dir, advanced_name):
        """
        Setup all the files required to enable an advanced theme. Copies
        all the files over and creates the required directories if they
        do not exist.
        :param theme_name: theme to copy the files over from
        :param output_dir: output directory to place the files in
        """
        """Directories"""
        output_theme_dir = os.path.join(output_dir, advanced_name)
        output_images_dir = os.path.join(output_theme_dir, advanced_name)
        input_theme_dir = os.path.join(utils.get_themes_directory(), theme_name)
        input_images_dir = os.path.join(input_theme_dir, theme_name)
        advanced_pkg_dir = os.path.join(utils.get_file_directory(), "advanced")
        """Directory creation"""
        for directory in [output_dir, output_theme_dir]:
            utils.create_directory(directory)
        """Theme TCL file"""
        file_name = theme_name + ".tcl"
        theme_input = os.path.join(input_theme_dir, file_name)
        theme_output = os.path.join(output_theme_dir, "{}.tcl".format(advanced_name))
        with open(theme_input, "r") as fi, open(theme_output, "w") as fo:
            for line in fi:
                # Setup new theme
                line = line.replace(theme_name, advanced_name)
                # Setup new image format
                line = line.replace("gif89", "png")
                line = line.replace("gif", "png")
                # Write processed line
                fo.write(line)
        """pkgIndex.tcl file"""
        theme_pkg_input = os.path.join(advanced_pkg_dir, "pkgIndex.tcl")
        theme_pkg_output = os.path.join(output_theme_dir, "pkgIndex.tcl")
        with open(theme_pkg_input, "r") as fi, open(theme_pkg_output, "w") as fo:
            for line in fi:
                fo.write(line.replace("advanced", advanced_name))
        """pkgIndex_package.tcl -> pkgIndex.tcl"""
        theme_pkg_input = os.path.join(advanced_pkg_dir, "pkgIndex_package.tcl")
        theme_pkg_output = os.path.join(output_dir, "pkgIndex.tcl")
        with open(theme_pkg_input, "r") as fi, open(theme_pkg_output, "w") as fo:
            for line in fi:
                fo.write(line.replace("advanced", advanced_name))
        """Images"""
        if os.path.exists(output_images_dir):
            rmtree(output_images_dir)
        copytree(input_images_dir, output_images_dir)

    @staticmethod
    def _setup_images(directory, brightness, saturation, hue, preserve_transparency):
        """
        Perform operations on the images found in the folder provided
        as a parameter. Used to modify images of the advanced theme.
        :param directory: absolute path to writable directory with images
        :param brightness: brightness modifier
        :param saturation: saturation modifier
        :param hue: hue modifier
        """
        for file_name in os.listdir(directory):
            with open(os.path.join(directory, file_name), "rb") as fi:
                image = Image.open(fi).convert("RGBA")
            # Only perform required operations
            if brightness != 1.0:
                enhancer = ImageEnhance.Brightness(image)
                image = enhancer.enhance(brightness)
            if saturation != 1.0:
                enhancer = ImageEnhance.Color(image)
                image = enhancer.enhance(saturation)
            if hue != 1.0:
                image = imgops.shift_hue(image, hue)
            if preserve_transparency is True:
                image = imgops.make_transparent(image)
            # Save the new image
            image.save(os.path.join(directory, file_name.replace("gif", "png")))
            image.close()
        for file_name in (item for item in os.listdir(directory) if item.endswith(".gif")):
            os.remove(os.path.join(directory, file_name))
        return

    def _load_tkimg(self):
        """
        Load the TkImg library from the tkimg folder for the current
        tk interpreter. Required for PNG support on Python 2, but also
        used on Python 3 as the theme files have been modified to
        support Img.
        :raise: tk.TclError if evaluation fails
        """
        tkimg_folder = utils.get_tkimg_directory()
        if not os.path.exists(tkimg_folder):
            raise RuntimeError("Unsupported platform and/or architecture")
        pkg_index = os.path.join(tkimg_folder, "pkgIndex.tcl")

        with utils.temporary_chdir(tkimg_folder):
            self.tk.call("lappend", "auto_path", "[{}]".format(tkimg_folder))
            self.tk.eval("source {}".format(os.path.relpath(pkg_index, os.getcwd())))
            self.tk.call("package", "require", "Img")
        return True

    @property
    def img_support(self):
        return self._img_support



"""
Author: RedFantom
License: GNU GPLv3
Copyright (c) 2017-2018 RedFantom
"""
from setuptools import setup
import os
import sys
from platform import architecture
from shutil import copytree, rmtree


platforms = {
    "win32": "win",
    "linux2": "linux"
}


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


if __name__ == '__main__':
    current_dir = os.path.dirname(__file__)
    dest_dir = os.path.join(current_dir, "ttkthemes", "tkimg")

    if os.path.exists(os.path.join(current_dir, "tkimg")):
        # ttkthemes source is in the ttkthemes/ folder and tkimg needs to
        # be copied there
        if "sdist" not in sys.argv[1]:
            # Set up TkImg library for bdist_wheel or bdist_wheel install
            prefix = sys.platform if sys.platform not in platforms else platforms[sys.platform]
            arch = int(architecture()[0][:2])
            tkimg_folder = os.path.join(current_dir, "tkimg", "{}{}".format(prefix, arch))
        else:
            # Set up TkImg library for sdist
            tkimg_folder = os.path.join(current_dir, "tkimg")

        if os.path.exists(dest_dir):
            rmtree(dest_dir)
        copytree(tkimg_folder, dest_dir)


setup(
    name='ttkthemes',
    packages=['ttkthemes'],
    package_data={"ttkthemes": ["themes/*", "tkimg/*"]},
    version='2.0.6',
    description='A group of themes for the ttk extensions of Tkinter with a Tkinter.Tk wrapper',
    author='The ttkthemes authors',
    author_email='redfantom@outlook.com',
    url='https://github.com/RedFantom/ttkthemes',
    download_url='https://github.com/RedFantom/ttkthemes/releases',
    include_package_data=True,
    keywords=['tkinter', 'ttk', 'gui', 'tcl', 'theme'],
    license='GPLv3',
    long_description=read('README.md'),
    classifiers=['Programming Language :: Python :: 2.7',
                 'Programming Language :: Python :: 3.4',
                 'Programming Language :: Python :: 3.5',
                 'Programming Language :: Python :: 3.6',
                 'Environment :: Win32 (MS Windows)',
                 'Environment :: X11 Applications',
                 'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
                 'Topic :: Software Development :: Libraries :: Tcl Extensions',
                 'Topic :: Software Development :: Libraries :: Python Modules'],
    zip_safe=False,
    install_requires=["pillow"],
    has_ext_modules=lambda: True,
)



"""
Author: RedFantom
License: GNU GPLv3
Copyright (c) 2017-2018 RedFantom
"""
import contextlib
import sys
import os
from shutil import rmtree
from platform import architecture
from tempfile import gettempdir


platforms = {
    "win32": "win",
    "linux2": "linux"
}


def is_python_3():
    """Check interpreter version"""
    return sys.version_info.major == 3


@contextlib.contextmanager
def temporary_chdir(new_dir):
    """
    Like os.chdir(), but always restores the old working directory.

    For example, code like this...

        old_curdir = os.getcwd()
        os.chdir('stuff')
        do_some_stuff()
        os.chdir(old_curdir)

    ...leaves the current working directory unchanged if do_some_stuff()
    raises an error, so it should be rewritten like this:

        old_curdir = os.getcwd()
        os.chdir('stuff')
        try:
            do_some_stuff()
        finally:
            os.chdir(old_curdir)

    Or equivalently, like this:

        with utils.temporary_chdir('stuff'):
            do_some_stuff()
    """
    old_dir = os.getcwd()
    os.chdir(new_dir)
    try:
        yield
    finally:
        os.chdir(old_dir)


def get_file_directory():
    """Return an absolute path to the current file directory"""
    return os.path.dirname(os.path.realpath(__file__))


def get_temp_directory():
    """Return an absolute path to an existing temporary directory"""
    # Supports all platforms supported by tempfile
    directory = os.path.join(gettempdir(), "ttkthemes")
    if not os.path.exists(directory):
        os.makedirs(directory)
    return directory


def get_themes_directory():
    """Return an absolute path the to /themes directory"""
    return os.path.join(get_file_directory(), "themes")


def create_directory(directory):
    """Create directory but first delete it if it exists"""
    if os.path.exists(directory):
        rmtree(directory)
    os.makedirs(directory)
    return directory


def get_tkimg_directory():
    """
    Return an absolute path to the TkImg directory for current platform
    """
    # Binary Distribution
    tkimg = os.path.join(os.path.dirname(__file__), "tkimg")
    if not os.path.exists(os.path.join(tkimg, "pkgIndex.tcl")):
        # Source Distribution
        prefix = sys.platform if sys.platform not in platforms else platforms[sys.platform]
        arch = architecture()[0][:2]
        tkimg = os.path.join(tkimg, "{}{}".format(prefix, arch))
        if not os.path.exists(tkimg):
            # Plain repository clone without having run bdist_wheel
            tkimg = os.path.realpath(
                os.path.join(
                    os.path.dirname(__file__),
                    "..",
                    "tkimg",
                    "{}{}".format(prefix, arch)))
    return tkimg
